from zodipy import emissivities
from zodipy import parameters
from zodipy._model import IPDModelRegistry
from zodipy._component_labels import ComponentLabel


model_registry = IPDModelRegistry()

model_registry.register_model(
    name="Planck13",
    components=[
        ComponentLabel.CLOUD,
        ComponentLabel.BAND1,
        ComponentLabel.BAND2,
        ComponentLabel.BAND3,
        ComponentLabel.RING,
        ComponentLabel.FEATURE,
    ],
    parameters=parameters.K98,
    emissivities=emissivities.PLANCK_2013,
)

model_registry.register_model(
    name="Planck15",
    components=[
        ComponentLabel.CLOUD,
        ComponentLabel.BAND1,
        ComponentLabel.BAND2,
        ComponentLabel.BAND3,
    ],
    parameters=parameters.K98,
    emissivities=emissivities.PLANCK_2015,
)

model_registry.register_model(
    name="Planck18",
    components=[
        ComponentLabel.CLOUD,
        ComponentLabel.BAND1,
        ComponentLabel.BAND2,
        ComponentLabel.BAND3,
    ],
    parameters=parameters.K98,
    emissivities=emissivities.PLANCK_2018,
)

model_registry.register_model(
    name="K98",
    components=[
        ComponentLabel.CLOUD,
        ComponentLabel.BAND1,
        ComponentLabel.BAND2,
        ComponentLabel.BAND3,
        ComponentLabel.RING,
        ComponentLabel.FEATURE,
    ],
    parameters=parameters.K98,
)



from __future__ import annotations
from dataclasses import dataclass, field
from typing import Dict, List, Optional

from zodipy._components import Component
from zodipy._component_labels import ComponentLabel, LABEL_TO_CLASS
from zodipy._emissivities import Emissivity


@dataclass
class IPDModel:
    """An initialized Interplanetary Dust Model.
    
    A IPDModel consists of a set of initialized `Components` and optionally
    a set of emissivities."""

    components: Dict[ComponentLabel, Component]
    emissivities: Optional[Emissivity] = None

    @property
    def includes_earth_neighboring_components(self) -> bool:
        """Returns True if the model includes a earth neighboring component."""

        return (
            ComponentLabel.RING in self.components
            or ComponentLabel.FEATURE in self.components
        )

    def __getitem__(self, component_name: str) -> Component:
        """Returns a sky component from the cosmoglobe model."""

        return self.components[ComponentLabel(component_name)]


@dataclass
class IPDModelRegistry:
    """Container for registered IPDModels."""

    registry: Dict[str, IPDModel] = field(default_factory=dict)

    def register_model(
        self,
        name: str,
        components: List[ComponentLabel],
        parameters: Dict[ComponentLabel, Dict[str, float]],
        emissivities: Optional[Emissivity] = None,
    ) -> None:
        """Adds a new IPD model to the registry."""

        if name in self.registry:
            raise ValueError(f"model by name {name} is already registered.")

        initialized_components: Dict[ComponentLabel, Component] = {}
        for component_label in components:
            component_class = LABEL_TO_CLASS[component_label]
            initialized_components[component_label] = component_class(
                **parameters[component_label]
            )

        self.registry[name] = IPDModel(
            components=initialized_components, emissivities=emissivities
        )

    def get_model(self, name: str) -> IPDModel:
        """Returns a IPDModel from the registry."""

        if name not in self.registry:
            raise ModuleNotFoundError(
                f"{name} is not a model in the registry. Avaliable models are "
                f"{', '.join(self.registry)}"
            )
            
        return self.registry[name]



from typing import List

import healpy as hp
import numpy as np

from zodipy._components import Component
from zodipy._integration import line_of_sight_integrate


def instantaneous_emission(
    nside: int,
    freq: float,
    components: List[Component],
    emissivities: List[float],
    line_of_sights: List[np.ndarray],
    observer_coords: np.ndarray,
    earth_coords: np.ndarray,
) -> np.ndarray:
    """Returns the simulated instantaneous emission.

    The emission is that seen by an observer at an instant at a single time
    or as the average of multiple times if several observer and Earth 
    coordinates are provided.

    Parameters:
    -----------
    nside
        HEALPIX map resolution parameter.
    freq:
        Frequency of the observer in GHz.
    components:
        Dictionary containing the Zodiacal Component that is used to evaluate
        the emission.
    emissivities:
        Sequency of emissivities, one for each component, corresponding to the
        frequency of `freq`.
    observer_coords
        Heliocentric coordinates of the observer.
    earth_coords
        Heliocentric coordinates of the Earth.
    """

    npix = hp.nside2npix(nside)
    unit_vectors = np.asarray(hp.pix2vec(nside, np.arange(npix)))

    emission = np.zeros((len(components), npix))
    for obs_pos, earth_pos in zip(
        observer_coords,
        earth_coords,
    ):
        for idx, component in enumerate(components):
            integrated_comp_emission = line_of_sight_integrate(
                line_of_sight=line_of_sights[idx],
                get_emission_func=component.get_emission,
                observer_coordinates=obs_pos,
                earth_coordinates=earth_pos,
                unit_vectors=unit_vectors,
                freq=freq,
            )

            emission[idx] += integrated_comp_emission * emissivities[idx]

    return emission * 1e20


def time_ordered_emission(
    nside: int,
    freq: float,
    components: List[Component],
    emissivities: List[float],
    line_of_sights: List[np.ndarray],
    observer_coordinates: np.ndarray,
    earth_coordinates: np.ndarray,
    pixel_chunk: np.ndarray,
    bin: bool = False,
) -> np.ndarray:
    """Simulates and returns the Zodiacal emission timestream.

    Parameters:
    -----------
    nside
        HEALPIX map resolution parameter.
    freq:
        Frequency of the observer in GHz.
    components:
        Dictionary containing the Zodiacal Component that is used to evaluate
        the emission.
    emissivities:
        Sequency of emissivities, one for each component, corresponding to the
        frequency of `freq`.
    observer_coords
        Heliocentric coordinates of the observer.
    earth_coords
        Heliocentric coordinates of the Earth.
    bin
        If True, return a binned HEALPIX map of the emission. If False, the
        timestream is returned.

    Returns
    -------
        Zodiacal emission [MJy/sr] over a timestream of pixels, or the
        binned Zodiacal emission map if bin is set to True.
    """

    if bin:
        pixels, counts = np.unique(pixel_chunk, return_counts=True)
        unit_vectors = np.asarray(hp.pix2vec(nside, pixels))
        emission = np.zeros((len(components), hp.nside2npix(nside)))

        for idx, component in enumerate(components):
            integrated_comp_emission = line_of_sight_integrate(
                line_of_sight=line_of_sights[idx],
                get_emission_func=component.get_emission,
                observer_coordinates=observer_coordinates,
                earth_coordinates=earth_coordinates,
                unit_vectors=unit_vectors,
                freq=freq,
            )
            emission[idx, pixels] = integrated_comp_emission * emissivities[idx]

        emission[:, pixels] *= counts

        return emission * 1e20

    pixels, indicies = np.unique(pixel_chunk, return_inverse=True)
    unit_vectors = np.asarray(hp.pix2vec(nside, pixels))
    time_stream = np.zeros((len(components), len(pixel_chunk)))

    for idx, component in enumerate(components):
        integrated_comp_emission = line_of_sight_integrate(
            line_of_sight=line_of_sights[idx],
            get_emission_func=component.get_emission,
            observer_coordinates=observer_coordinates,
            earth_coordinates=earth_coordinates,
            unit_vectors=unit_vectors,
            freq=freq,
        )

        time_stream[idx] = integrated_comp_emission[indicies] * emissivities[idx]

    return time_stream * 1e20



from dataclasses import dataclass
from typing import Dict, List, Optional, Sequence

import numpy as np

from zodipy._component_labels import ComponentLabel


@dataclass
class Emissivity:
    """Emissivity parameters for a Zodiacal component."""

    frequencies: Sequence[float]
    components: Dict[ComponentLabel, Sequence[float]]

    def __call__(self, component: ComponentLabel, freq: float) -> float:
        """Returns the interpolated emissivity given a frequency.

        Parameters
        ----------
        component
            Component label.
        freq
            Frequency at which to evaluate the Zodiacal emission.

        Returns
        -------
        emissivity
            Interpolated emissivity scaling factor.
        """

        if not self.frequencies[0] <= freq <= self.frequencies[-1]:
            raise ValueError(f"Frequency is out of range")

        emissivity = np.interp(freq, self.frequencies, self.components[component])

        return emissivity


def get_emissivities(
    freq: float,
    emissivity: Optional[Emissivity],
    components: List[ComponentLabel],
) -> List[float]:
    """Returns a list of interpolated emissivities for each component."""

    if emissivity is not None:
        return [emissivity(component, freq) for component in components]

    return [1.0 for _ in components]



from typing import Optional, Sequence, Union, Dict

from astroquery.jplhorizons import Horizons
import healpy as hp
import numpy as np

from zodipy._exceptions import TargetNotSupportedError

EpochsType = Optional[Union[float, Sequence[float], Dict[str, str]]]


TARGET_IDS = {
    "planck": "Planck",
    "wmap": "WMAP",
    "l2": "32",
    "sun": "10",
    "mercury": "199",
    "venus": "299",
    "earth": "399",
    "mars": "499",
    "jupiter": "599",
}


def get_target_coordinates(
    target: str,
    epochs: Optional[EpochsType] = None,
) -> np.ndarray:
    """Returns the heliocentric coordinates of the target given an epoch.

    Parameters
    ----------
    target
        Name of the target body.
    epochs
        Either a list of epochs in JD or MJD format or a dictionary
        defining a range of times and dates; the range dictionary has to
        be of the form {``'start'``:'YYYY-MM-DD [HH:MM:SS]',
        ``'stop'``:'YYYY-MM-DD [HH:MM:SS]', ``'step'``:'n[y|d|h|m|s]'}.
        If no epochs are provided, the current time is used.
    return_dates
        Boolean for wheter or not to returns the Julian dates for each
        target location.

    Returns
    -------
    coordinates
        Heliocentric cartesian coordinates of the target.
    dates, optional
        Julian dates of each target coordinate.
    """

    try:
        target = TARGET_IDS[target.lower()]
    except KeyError:
        raise TargetNotSupportedError(
            f"{target} is not a valid observer. Please select one of the following: "
            f"{', '.join(TARGET_IDS)}."
        )

    if epochs is None:
        epochs = [2459215.50000]  # 01-01-2021

    query = Horizons(id=target, id_type="majorbody", location="c@sun", epochs=epochs)
    ephemerides = query.ephemerides()

    R = ephemerides["r"].value
    lon, lat = ephemerides["EclLon"].value, ephemerides["EclLat"].value
    lon, lat = np.deg2rad(lon), np.deg2rad(lat)

    x = R * np.cos(lat) * np.cos(lon)
    y = R * np.cos(lat) * np.sin(lon)
    z = R * np.sin(lat)

    coordinates = np.stack((x, y, z), axis=1)

    return coordinates


def to_frame(input_map: np.ndarray, coord_out: str, coord_in: str = "E") -> np.ndarray:
    """Rotates a HEALPIX map to another reference frame.

    Parameters
    ----------
    input_map
        Map to rotate.
    coord_out
        Coordinate system of the output_map.
    coord_in
        Coordinate system of the input_map.

    Returns
    -------
    output_map
        Rotated map.
    """

    if coord_in == coord_out:
        return input_map

    nside = hp.get_nside(input_map)
    npix = hp.nside2npix(nside)

    rotator = hp.Rotator(coord=[coord_out, coord_in])
    new_angles = rotator(hp.pix2ang(nside, np.arange(npix)))
    new_pixels = hp.ang2pix(nside, *new_angles)
    output_map = input_map[..., new_pixels]

    return output_map



from typing import Dict, Optional, Union

import numpy as np

from zodipy._coordinates import to_frame, get_target_coordinates, EpochsType
from zodipy._emissivities import get_emissivities
from zodipy._integration_config import integration_config_registry
from zodipy._simulation import instantaneous_emission, time_ordered_emission
from zodipy.models import model_registry


class InterplanetaryDustModel:
    """The Zodipy simulation interface."""

    def __init__(self, model: str = "K98") -> None:
        """Initializes the interface given an interplaneteray dust model.

        By default, the Interplanetary Dust Model used by Zodipy is the
        Kelsall et al. (1998) Interplanetary Dust Model, which includes
        five Zodiacal components:
            - The Diffuse Cloud
            - Three Asteroidal Bands
            - The Circumsolar Ring
            - The Earth-trailing Feature

        The Kelsall model yields the Zodiacal Emission given purely by the
        parametric model, and does not use the emissivity factors fitted by
        the Planck Collaboration.

        Other available models are:
            - Planck13 (all five Kelsall components + emissivity fits for each)
            - Planck15 (cloud + bands + new emissivity fits)
            - Planck18 (cloud + bands + the latest emissivity fits)

        Parameters
        ----------
        model
            The name of a implemented Interplanetary Dust Model.
        """

        self.model = model_registry.get_model(model)
        integration_config = integration_config_registry.get_config("default")
        self.line_of_sights = [
            line_of_sight for line_of_sight in integration_config.values()
        ]

    def get_instantaneous_emission(
        self,
        nside: int,
        freq: float,
        observer: str = "L2",
        epochs: Optional[EpochsType] = None,
        coord: str = "E",
        return_comps: bool = False,
    ) -> Union[np.ndarray, Dict[str, np.ndarray]]:
        """Simulates and returns the instantaneous Zodiacal Emission [MJy/sr].

        By instantaneous emission we mean the emission observed in an instant
        in time.

        The observer location (and optionally the location of the Earth if
        either of the Feature or the Ring components are included) are queried
        from the Horizons JPL ephemerides, given some epoch.

        NOTE: This function returns the fullsky emission from one coordinate
        in space. This means that we evaluate line-of-sights looking in towards
        the inner Solar System where the dust density increases exponentially.
        These are line-of-sights unlikely to be observed by an actual
        experiment.

        Parameters
        ----------
        nside
            HEALPIX map resolution parameter.
        freq
            Frequency at which to evaluate the Zodiacal emission in units of GHz.
        observer
            The observer. Defaults to 'L2'.
        epochs
            The observeration times given as a single epoch, or a list of epochs
            in JD or MJD format, or a dictionary defining a range of times and
            dates; the range dictionary has to be of the form
            {'start':'YYYY-MM-DD [HH:MM:SS]', 'stop':'YYYY-MM-DD [HH:MM:SS]',
            'step':'n[y|d|h|m|s]'}. If no epochs are provided, the current time
            is used in UTC.
        coord
            Coordinate frame of the output map. Defaults to 'E', which is
            ecliptic coordinates.
        return_comps
            If True, the emission of each component in the model is
            returned individually in a dictionary. Defaults to False.

        Returns
        -------
        emission
            Simulated instantaneous Zodiacal emission in units of MJy/sr.
        """

        observer_coordinates = get_target_coordinates(observer, epochs)
        if self.model.includes_earth_neighboring_components:
            earth_coordinates = get_target_coordinates("earth", epochs)
        else:
            earth_coordinates = np.zeros(3)

        emissivities = get_emissivities(
            freq=freq,
            emissivity=self.model.emissivities,
            components=list(self.model.components.keys()),
        )

        emission = instantaneous_emission(
            nside=nside,
            freq=freq,
            components=list(self.model.components.values()),
            emissivities=emissivities,
            observer_coords=observer_coordinates,
            earth_coords=earth_coordinates,
            line_of_sights=self.line_of_sights,
        )
        if coord != "E":
            emission = to_frame(emission, coord)

        if not return_comps:
            return emission.sum(axis=0)

        return {
            comp.value: emission[idx] for idx, comp in enumerate(self.model.components)
        }

    def get_time_ordered_emission(
        self,
        nside: int,
        freq: float,
        pixels: np.ndarray,
        observer_coordinates: np.ndarray,
        earth_coordinates: Optional[np.ndarray] = None,
        return_comps: bool = False,
        bin: bool = False,
    ) -> Union[np.ndarray, Dict[str, np.ndarray]]:
        """Simulates and returns the Zodiacal emission timestream [MJy/sr].

        Given a timestream of pixels, this function evaluates the Zodiacal
        Emission given a position in space (`observer_coordinates`), and
        optionally Earth's coordinates (`earth_coordinates`) if the
        interplanetary dust model includes either of the Feature or Ring
        components.

        If the bin parameter is set to True, the timestream is binned into a
        HEALPIX map instead.

        Parameters
        ----------
        nside
            HEALPIX map resolution parameter.
        freq
            Frequency at which to evaluate the Zodiacal emission in units of
            GHz.
        pixels
            Chunk of time-ordered pixels corresponding to a parts of a scanning
            strategy.
        observer_coordinates
            The heliocentric coordinates with shape (3,) of the observer over
            the tod chunk.
        earth_coordinates
            The heliocentric coordinates with shape (3,) of the Earth over the
            tod chunk.
        return_comps
            If True, the emission of each component in the model is
            returned individually in a dictionary. Defaults to False.
        bin
            If True, return a binned HEALPIX map of the emission. If False, the
            timestream is returned.

        Returns
        -------
        emission
            Zodiacal emission [MJy/sr] over a timestream of pixels, or the
            binned Zodiacal emission map if bin is set to True.
        """

        if earth_coordinates is None:
            if self.model.includes_earth_neighboring_components:
                raise ValueError(
                    "The Interplanetary Dust model includes earth-neighboring "
                    "components, but no earth_coordinates were given."
                )
            earth_coordinates = np.zeros(3)

        emissivities = get_emissivities(
            freq=freq,
            emissivity=self.model.emissivities,
            components=list(self.model.components.keys()),
        )

        emission = time_ordered_emission(
            nside=nside,
            freq=freq,
            components=list(self.model.components.values()),
            emissivities=emissivities,
            line_of_sights=self.line_of_sights,
            observer_coordinates=observer_coordinates,
            earth_coordinates=earth_coordinates,
            pixel_chunk=pixels,
            bin=bin,
        )

        if not return_comps:
            return emission.sum(axis=0)

        return {
            comp.value: emission[idx] for idx, comp in enumerate(self.model.components)
        }